/*************************************************************************
Title:    GPS-Empfnger
Author:   Thomas Wiens
Date:     5.8.2006
File:     main.c
Compiler: AVR-GCC 3.4.3
Hardware: HD44780 compatible LCD text display
          ATMega8
 
**************************************************************************/
// Die Buffer Bibliothek wurde aus der Avrlib entnommen. Die GPS-Funktionen 
// von Gleitpunkt auf normale Strings umgendert, da mit den Werten nicht gerechnet
// werden soll.

// Bedienung:
//  Beim Einschalten:
//      Start-Taste gedrckt:   Wechselt in den bertragungsmodus
//      Stop-Taste gedrckt:    Beginnt einen neuen Track
//  Im Betrieb:
//      Start-Taste:            Startet das Logging
//      Stop-Taste:             Stoppt das Logging
// Vor dem Ausschalten des Gertes sollte der Logging Vorgang gestoppt werden.
// Damit wird der aktuelle Index im EEPROM gespeichert, und der Track kann
// beim nchsten Einschalten fortgesetzt werden.

#include "main.h"

// #define LCD_DISPLAY

#define UART_RX_BUFFER_SIZE     0x00A0      // 160 Zeichen
#define DEBOUNCE                256L        // debounce clock (256Hz = 4msec)
#define LOGGING_SIZE            2000        // Speichergre Positionsdatenstze externes EEPROM
#define MAX_WAYPOINTS           200         // Anzahl der speicherbaren Wegpunkte internes EEPROM

#define MODE_INACTIVE           0x00        // Gert inaktiv, Logging nicht aktiv
#define MODE_LOGGING            0x01        // Gert speichert GPS-Daten
#define MODE_MEM_FULL           0x02        // Im Logging Mode Speicher voll
#define MODE_AC_EMPTY           0x04        // Akku fast leer
#define MODE_TRANSMIT           0x08        // Gert im bertragungsmodus
#define MODE_CLEAR              0x10        // Datensatzlschmodus
#define MODE_Q_BAD              0x20        // GPS-Empfang schlecht
#define MODE_STORE_INDEX        0x40        // Aktuellen Index im EEPROM sichern
#define MODE_STORE_WAYPT        0x80        // Aktuellen Punkt speichern und Index als Wegpunkt sichern
//************************************************************************************************************
// PROTOTYPES
//************************************************************************************************************
void init_io(void);
void init_timer(void);
uint8_t store_gps_position(void);
void store_waypoint(void);
void write_gps_to_lcd(void);
// void dump_eeprom(void);
void delay_s(uint8_t sec);
//************************************************************************************************************
// GLOBAL VARIABLES
//************************************************************************************************************
uint8_t Prescaler;
uint8_t Key_old;
uint8_t volatile Mode;
uint8_t volatile Second;                                        // count Seconds
uint16_t volatile StoreSecond;                                  // Sekundenzhler fr GPS-Logging Zyklus
uint16_t StoreCycle;                                             // Eingestellter Logging_Zyklus in s

cBuffer UartRxBuffer;                                       // uart receive buffer
uint16_t Eeprom_position = 0;
uint16_t NumWaypoints = 0;

struct GpsInfoType GpsInfo;
static char uartRxData[UART_RX_BUFFER_SIZE];

//************************************************************************************************************
// EEPROM VARIABLES
//************************************************************************************************************
uint16_t ee_eeprom_position EEPROM = 0;                          // Letzter Speicherindex
uint16_t eeStoreCycle EEPROM = 10;                               // Im EEPROM gespeicherter Logging Zyklus in s

uint16_t eeNumWaypoints EEPROM = 0;                              // Anzahl der gespeicherten Wegpunkte
uint16_t eeWaypoint[MAX_WAYPOINTS] EEPROM;                       // Speicher fr Wegpunkte im EEPROM des AVR 

//************************************************************************************************************
// INTERRUPTS
//************************************************************************************************************
SIGNAL (SIG_UART_RECV) {    
    char zeichen;
    zeichen = UDR;
    bufferAddToEnd(&UartRxBuffer, zeichen);
}
//************************************************************************************************************
SIGNAL (SIG_OUTPUT_COMPARE1A) {
    uint8_t tcnt1h = TCNT1H;
    OCR1A += XTAL / DEBOUNCE;                               // new compare value
    if( ++Prescaler == (uint8_t)DEBOUNCE ) {
        Prescaler = 0;
        Second++;                                           // exact one Second over
        StoreSecond++;
        #if XTAL % DEBOUNCE                                 // handle remainder
            OCR1A += XTAL % DEBOUNCE;                       // compare once per Second
        #endif
        // Switch LEDs       
        if (Mode & MODE_TRANSMIT)
            LED_GREEN_TOOGLE;
        else if (Mode & MODE_LOGGING)
            LED_GREEN_ON;
        if (Mode & (MODE_Q_BAD) || Mode & (MODE_MEM_FULL))
            LED_RED_ON;        
    }
    else {                                                  // all LEDs will be switched when the next interrupt occurs
        if (!(Mode & MODE_TRANSMIT)) {
            if (Mode & MODE_Q_BAD || Mode & MODE_LOGGING) {
                LED_GREEN_OFF;
                LED_RED_OFF;
            }
        }
    }
    TCNT1H = tcnt1h;                                        // restore for delay() !    

    if(!(PIN_SW_START & (1 << SW_START)) && Key_old && (Mode & MODE_LOGGING)) {		    // rising edge
        Mode |= MODE_STORE_WAYPT;
    }
    
    if (!(Mode & MODE_TRANSMIT)) {                        // Check keys only if not in transmit-mode
        // Start pressed
        if (!(PIN_SW_START & (1 << SW_START))) {
            Mode |= MODE_LOGGING;
        }
        // Stop pressed
        else if (!(PIN_SW_STOP & (1 << SW_STOP)) && (Mode & MODE_LOGGING)) {
            Mode &= ~(MODE_LOGGING);
            Mode |= (MODE_STORE_INDEX);
        }
    }
    Key_old = (PIN_SW_START & (1 << SW_START));             // Save last key-state
}

//************************************************************************************************************
// MAIN
//************************************************************************************************************
int main(void) {
    init_io();
    init_twi();
    LED_GREEN_OFF;
    LED_RED_OFF;
    Mode = MODE_INACTIVE;
    init_timer();
    #ifdef LCD_DISPLAY
    lcd_init(LCD_DISP_ON);
    lcd_clrscr();
    lcd_gotoxy(0,4);
    lcd_puts("GPS-Empfaenger\n");
    #endif
    Second = 0;   
    StoreSecond = 0;
    
    StoreCycle = eeprom_read_word(&eeStoreCycle);                     // Read logging cycle from internal EEPROM
    Eeprom_position = eeprom_read_word(&ee_eeprom_position);          // Read last index to continue logging
    NumWaypoints = eeprom_read_word(&eeNumWaypoints);
    
    bufferInit(&UartRxBuffer, uartRxData, UART_RX_BUFFER_SIZE);       
    
    // Startup-Mode
    if (!(PIN_SW_START & (1 << SW_START)))
        Mode |= MODE_TRANSMIT;
    else if (!(PIN_SW_STOP & (1 << SW_STOP))) {
        // Start new track
        Eeprom_position = 0;
        NumWaypoints = 0;
        eeprom_write_word(&ee_eeprom_position, Eeprom_position);         
        eeprom_write_word(&eeNumWaypoints, NumWaypoints); 
        LED_RED_ON;
        delay_s(1);
        LED_RED_OFF;
        LED_GREEN_ON;
        delay_s(1);
        LED_GREEN_OFF;
    }
    // Init UART and set baudrate    
    if (Mode & MODE_TRANSMIT) {
        MAX232_ON;
        uinit(B_4800);                                              // Baudrate for PC-Communication
    }
    else {
        MAX232_OFF;
        uinit(B_4800);                                               // 4800 Baud for NMEA
    }    
    UCSRB |= (1 << RXCIE);                                          // Enable receive interrupt        
    sei();    
    for (;;) {
        while(nmeaProcess(&UartRxBuffer));     
        
        if (!(Mode & MODE_TRANSMIT)) {                
            if (GpsInfo.fix[0] < '2')                              // 2d-fix not available
                Mode |= MODE_Q_BAD;
            else
                Mode &= ~(MODE_Q_BAD);                
            if( Second >= 2){
            #ifdef LCD_DISPLAY
                write_gps_to_lcd();
            #endif
                Second = 0;
            }
            if (StoreSecond >= StoreCycle) {    
                if (Mode & MODE_LOGGING)
                    store_gps_position();
                StoreSecond = 0;
            }
            if (Mode & MODE_STORE_WAYPT) {
                store_gps_position();
                store_waypoint();
                Mode &= ~(MODE_STORE_WAYPT);
            }    
        }
        if (Mode & MODE_STORE_INDEX) {
            eeprom_write_word(&ee_eeprom_position, Eeprom_position); 
            Mode &= ~(MODE_STORE_INDEX);
        }   
    }
    return 0;
}

//************************************************************************************************************
// FUNCTIONS
//************************************************************************************************************
#ifdef LCD_DISPLAY
void write_gps_to_lcd(void) {
    char buf[10];
    lcd_clrscr();
    lcd_puts(GpsInfo.utc_time);
    lcd_puts("-");
    lcd_puts(GpsInfo.utc_date);            
    lcd_puts("\n");
    lcd_puts("Laenge: ");            
    lcd_puts(GpsInfo.latitude);
    lcd_puts("\n");
    lcd_puts("Breite: ");
    lcd_puts(GpsInfo.longitude);
    lcd_puts("\n");
    // lcd_puts("H:");
    // lcd_puts(GpsInfo.altitude);
    lcd_puts("S:");
    itoa(StoreCycle, buf, 10);
    lcd_puts(buf);
    lcd_puts(" L:");
    itoa(Eeprom_position, buf, 10);
    lcd_puts(buf);
    lcd_puts(" M:");
    itoa(Mode, buf, 10);
    lcd_puts(buf);
    lcd_puts(" Q:");
    lcd_puts(GpsInfo.fix);
}
#endif
//************************************************************************************************************
void init_io(void) {
    DDR_LED_GREEN |= (1 << LED_GREEN);
    DDR_LED_RED |= (1 << LED_RED);
    DDR_MAX232 |= (1 << MAX232);
    // Swtiches as inputs with internal Pull-ups on
    DDR_SW_START &= ~(1 << SW_START);
    PORT_SW_START |= (1 << SW_START);

    DDR_SW_STOP &= ~(1 << SW_STOP);
    PORT_SW_STOP |= (1 << SW_STOP);    
    
    // Enable internal Pull-Ups for TWI-Interface
    DDRC &= ~((1 << PC4) | (1 << PC5));
    PORTC |= (1 << PC4) | (1 << PC5);    
}

//************************************************************************************************************
void init_timer(void) {
    TCCR1B = 1 << CS10;                                     // divide by 1
    OCR1A = 0;
    TCNT1 = -1;
    Second = 0;
    Prescaler = 0;
    TIMSK = 1 << OCIE1A;
}

//************************************************************************************************************
uint8_t store_gps_position(void) {
    tPositionSet Pos;
    uint16_t addr;
    char temp[12];
    
    if (Eeprom_position < LOGGING_SIZE) {
        Pos.date = atol(GpsInfo.utc_date);
        // Time-Format: "121630.611", but we don't want to know milliseconds        
        temp[0] = 0;
        strncpy(temp, GpsInfo.utc_time, 6);
        temp[6] = 0;
        Pos.time = atol(temp);    
        // Store latitude and langitude without post decimal positions
        // Latitude-Format: "5215.3472", dot at position 4        
        temp[0] = 0;
        strncpy(temp, GpsInfo.latitude, 4);
        temp[4] = 0;
        strcat(temp, &GpsInfo.latitude[5]);
        Pos.latitude = atol(temp);        
        // Altitude-Format: "00803.1681", dot at position 5       
        temp[0] = 0;
        strncpy(temp, GpsInfo.longitude, 5);
        temp[5] = 0;
        strcat(temp, &GpsInfo.longitude[6]);       
        Pos.longitude = atol(temp);
        // Store altitude without post decimal positions
        strcpy(temp, GpsInfo.altitude);
        temp[strlen(temp) - 2] = '\0';
        Pos.altitude = atoi(temp);        
        // Store direction as single char
        Pos.latitude_ns = GpsInfo.latitude_ns[0];
        Pos.longitude_ew = GpsInfo.longitude_ew[0];
        cli();        
        // Startadresse
        addr = sizeof (tPositionSet) * Eeprom_position;
        ee24xx_write_bytes(addr, sizeof (tPositionSet), (uint8_t*) &Pos);                
        // Flush UART-Buffer if characters are received during eeprom writing
        bufferFlush(&UartRxBuffer);
        sei();    
        Eeprom_position++;
        return 0;
    }
    else {
        Mode |= MODE_MEM_FULL; 
        return 1;
    }
}

void store_waypoint(void) {
    if (NumWaypoints >= MAX_WAYPOINTS || Eeprom_position == 0)
        return;        
    cli();       
    eeprom_write_word(&eeWaypoint[NumWaypoints], Eeprom_position - 1);      // position was incremented in gps_store funtcion    
    NumWaypoints++;    
    eeprom_write_word(&eeNumWaypoints, NumWaypoints);   
    LED_GREEN_ON;
    LED_RED_ON;
    delay_s(1);
    LED_GREEN_OFF;
    LED_RED_OFF;
    delay_s(1);
    LED_GREEN_ON;
    LED_RED_ON;
    delay_s(1);
    LED_GREEN_OFF;
    LED_RED_OFF;
    // Flush UART-Buffer if characters are received during eeprom writing
    bufferFlush(&UartRxBuffer);
    sei();  
}

void dump_eeprom(void) {
    uint16_t a;
    int rv;
    uint8_t b[16];
    uint8_t x;
    char buf[20];   
    for (a = 0; a < 1024;) {
        sprintf(buf, "%#04x:",  a);
        uputs(buf);
        rv = ee24xx_read_bytes(a, 16, b);
        if (rv <= 0)
            error();
        if (rv < 16) {
            sprintf(buf, "warning: short read %d\n",  rv);
            uputs(buf);
        }
        a += rv;
        for (x = 0; x < rv; x++) {
            sprintf(buf, "%02x ", b[x]);
            uputs(buf);            
        }
        uputs("\n");
    }
}

void delay_s(uint8_t sec) {
    uint16_t i;
    for (i = 0; i < (10 * sec); i++)
        _delay_ms(100.0);
}
