/*! \file nmea.c \brief NMEA protocol function library. */
//*****************************************************************************
//
// File Name	: 'nmea.c'
// Title		: NMEA protocol function library
// Author		: Pascal Stang - Copyright (C) 2002
// Created		: 2002.08.27
// Revised		: 2002.08.27
// Version		: 0.1
// Target MCU	: Atmel AVR Series
// Editor Tabs	: 4
//
// Changed datatypes to chars
// Thomas Wiens - August 2006
//*****************************************************************************

#ifndef WIN32
	#include <avr/io.h>
	#include <avr/signal.h>
	#include <avr/interrupt.h>
	#include <avr/pgmspace.h>
#endif
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "global.h"
#include "buffer.h"
#include "nmea_char.h"

// Global variables
extern struct GpsInfoType GpsInfo;
extern uint16_t Eeprom_position;
extern tPositionSet eePosition[];
extern cBuffer UartRxBuffer;				// uart receive buffer
extern uint16_t NumWaypoints;
extern uint16_t eeWaypoint[];
uint8_t NmeaPacket[NMEA_BUFFERSIZE];

void nmeaInit(void)
{
}

//************************************************************************************************************
uint8_t* nmeaGetPacketBuffer(void)
{
	return NmeaPacket;
}

//************************************************************************************************************
uint8_t nmeaProcess(cBuffer* rxBuffer) {
	uint8_t foundpacket = NMEA_NODATA;
	uint8_t startFlag = FALSE;
	//uint8_t data;
	uint16_t i,j;

	// process the receive buffer
	// go through buffer looking for packets
	while(rxBuffer->datalength)
	{
		// look for a start of NMEA packet
		if(bufferGetAtIndex(rxBuffer,0) == '$')
		{
			// found start
			startFlag = TRUE;
			// when start is found, we leave it intact in the receive buffer
			// in case the full NMEA string is not completely received.  The
			// start will be detected in the next nmeaProcess iteration.

			// done looking for start
			break;
		}
		else
			bufferGetFromFront(rxBuffer);
	}
	
	// if we detected a start, look for end of packet
	if(startFlag)
	{   
		for(i=1; i<(rxBuffer->datalength)-1; i++)
		{
			// check for end of NMEA packet <CR><LF>
			if((bufferGetAtIndex(rxBuffer,i) == '\r') && (bufferGetAtIndex(rxBuffer,i+1) == '\n'))
			{
				// have a packet end
				// dump initial '$'
				bufferGetFromFront(rxBuffer);
				// copy packet to NmeaPacket
				for(j=0; j<(i-1); j++)
				{
					// although NMEA strings should be 80 characters or less,
					// receive buffer errors can generate erroneous packets.
					// Protect against packet buffer overflow
					if(j<(NMEA_BUFFERSIZE-1))
						NmeaPacket[j] = bufferGetFromFront(rxBuffer);
					else
						bufferGetFromFront(rxBuffer);
				}
				// null terminate it
				NmeaPacket[j] = 0;
				// dump <CR><LF> from rxBuffer
				bufferGetFromFront(rxBuffer);
				bufferGetFromFront(rxBuffer);

				// found a packet
				// done with this processing session
				foundpacket = NMEA_UNKNOWN;
				break;
			}
		}
	}
    
	if(foundpacket)
	{   
		// check message type and process appropriately
		if(!strncmp(NmeaPacket, "GPGGA", 5))
		{
			// process packet of this type
			nmeaProcessGPGGA(NmeaPacket);
			// report packet type
			foundpacket = NMEA_GPGGA;
		}
		else if(!strncmp(NmeaPacket, "GPRMC", 5))
		{
			// process packet of this type
			nmeaProcessGPRMC(NmeaPacket);
			// report packet type
			foundpacket = NMEA_GPRMC;
		}
		else if(!strncmp(NmeaPacket, "GPGSA", 5))
		{
			// process packet of this type
			nmeaProcessGPGSA(NmeaPacket);
			// report packet type
			foundpacket = NMEA_GPGSA;
		}        
        // ****************** Betrieb am PC **************************************
        else if(!strncmp(NmeaPacket, "TREQ", 4))
		{
			pcTransTREQ();
            foundpacket = PC_TREQ;
		}
        else if(!strncmp(NmeaPacket, "TSET", 4))
		{
            pcTransTSET(NmeaPacket);
			foundpacket = PC_TSET;
		}
        else if(!strncmp(NmeaPacket, "WREQ", 4))
		{
            pcTransWREQ();
			foundpacket = PC_WREQ;
		}
        else if(!strncmp(NmeaPacket, "WSET", 4))
		{
            pcTransWSET(NmeaPacket);
			foundpacket = PC_WSET;
		}
        else if(!strncmp(NmeaPacket, "CLEAR", 5))
		{
            pcTransCLEAR();
			foundpacket = PC_CLEAR;
		}        
        
        // ************************************************************************
	}
	else if(rxBuffer->datalength >= rxBuffer->size)
	{
		// if we found no packet, and the buffer is full
		// we're logjammed, flush entire buffer
		bufferFlush(rxBuffer);
	}
	return foundpacket;
}

//************************************************************************************************************
/*
    $GPGGA,170834,4124.8963,N,08151.6838,W,1,05,1.5,280.2,M,-34.0,M,,,*75 
    
1       Time                                        170834          17:08:34 UTC 
2, 3    Latitude                                    4124.8963,  N   41d 24.8963' N or 41d 24' 54" N 
4, 5    Longitude                                   08151.6838, W   81d 51.6838' W or 81d 51' 41" W 
6       Fix Quality:
        - 0 = Invalid
        - 1 = GPS fix
        - 2 = DGPS fix                              1       Data is from a GPS fix 
7       Number of Satellites                        05      5 Satellites are in view 
8       Horizontal Dilution of Precision (HDOP)     1.5     Relative accuracy of horizontal position 
9       Altitude                                    280.2, M 280.2 meters above mean sea level 
10      Height of geoid above WGS84 ellipsoid       -34.0, M -34.0 meters 
11      Time since last DGPS update                 blank   No last update 
12      DGPS reference station id                   blank   No station id 

*/
void nmeaProcessGPGGA(uint8_t* packet) {
    uint16_t i, len;    
    char *endptr;
    char field;
    char buf[MAX];

	// start parsing just after "GPGGA,"
	i = 6;
	for (field = 1; field < 13; field++) {        
        endptr = strchr(&packet[i], ',');
        len = (int)endptr - (int)&packet[i];
        
        strncpy(buf, &packet[i], len);
        buf[len] = '\0';
        switch (field) {
            case 2:
                strcpy(GpsInfo.latitude, buf);
                break;
            case 3:
                strcpy(GpsInfo.latitude_ns, buf);
                break;
            case 4:
                strcpy(GpsInfo.longitude, buf);
                break;
            case 5:
                strcpy(GpsInfo.longitude_ew, buf);
                break;
            case 9:
                strcpy(GpsInfo.altitude, buf);             
        }
        buf[0] = '\0';
    	while(packet[i++] != ',');				// next field
    }   
}

//************************************************************************************************************
/*
    for NMEA 0183 version 3.00 active the Mode indicator field is added
    $GPRMC,hhmmss.ss,A,llll.ll,a,yyyyy.yy,a,x.x,x.x,ddmmyy,x.x,a,m*hh
Field #
1    = UTC time of fix
2    = Data status (A=Valid position, V=navigation receiver warning)
3    = Latitude of fix
4    = N or S of longitude
5    = Longitude of fix
6    = E or W of longitude
7    = Speed over ground in knots
8    = Track made good in degrees True
9    = UTC date of fix
10   = Magnetic variation degrees (Easterly var. subtracts from true course)
11   = E or W of magnetic variation
12   = Mode indicator, (A=Autonomous, D=Differential, E=Estimated, N=Data not valid)
13   = Checksum
*/
void nmeaProcessGPRMC(uint8_t* packet) {
    uint16_t i, len;
    char *endptr;
    char field;
    char buf[MAX];
	// start parsing just after "GPRMC,"
	i = 6;
	for (field = 1; field < 12; field++) {        
        endptr = strchr(&packet[i], ',');
        len = (int)endptr - (int)&packet[i];        // Hier Fehlfunktion, wenn & fehlt
        strncpy(buf, &packet[i], len);
        buf[len] = '\0';
        switch (field) {
            case 1:     // UTC-Uhrzeit
                strcpy(GpsInfo.utc_time, buf);
                break;
            case 9:     // UTC-Datum
                strcpy(GpsInfo.utc_date, buf);
                break;               
        }
        buf[0] = '\0';
    	while(packet[i++] != ',');				// next field
    }   
}

//************************************************************************************************************
/*
$GPGSA,A,3,,,,15,17,18,23,,,,,,4.7,4.4,1.5*3F
       ^ ^ ^                   ^   ^   ^
       | | |                   |   |   |
       | | |                   |   |   VDOP (vertikale Genauigkeit)
       | | |                   |   |
       | | |                   |   HDOP (horizontale Genauigkeit)
       | | |                   |
       | | |                   PDOP (Genauigkeit)
       | | |
       | | PRN-Nummern von maximal 12 Satelliten
       | |
       | Art der Positionsbestimmung (3 = 3D-fix)
       |                             (2 = 2D-fix)
       |                             (1 = kein Fix)
       |
       Auto-Auswahl 2D oder 3D Bestimmung
*/
void nmeaProcessGPGSA(uint8_t* packet) {
    uint16_t i, len;
    char *endptr;
    char field;
    char buf[MAX];
	// start parsing just after "GPGSA,"
	i = 6;
	for (field = 1; field < 17; field++) {        
        endptr = strchr(&packet[i], ',');
        len = (int)endptr - (int)&packet[i];        // Hier Fehlfunktion, wenn & fehlt
        strncpy(buf, &packet[i], len);
        buf[len] = '\0';
        switch (field) {
            case 2:     // Fix
                strcpy(GpsInfo.fix, buf);
                break;            
        }
        buf[0] = '\0';
    	while(packet[i++] != ',');				// next field
    }   
}
//************************************************************************************************************
// Funktionen zur bertragung der Datenstze zum PC
void pcTransTREQ(void) {
    char buf[10];
    itoa (Eeprom_position, buf, 10);
    uputs("$TCNT,");
    uputs(buf);
    uputs("*\r\n");    
}
//************************************************************************************************************
void pcTransTSET(uint8_t* packet) {
    tPositionSet Pos;
    uint16_t set_number, len;
    uint16_t addr;
    char *endptr;
    char buf[15];    
    // Interrupts aus
    cli();
    // Buffer leeren, da ansonsten mehrere Befehle mehrmals abgearbeitet werden
    bufferFlush(&UartRxBuffer);
    
    buf[0] = 0;
    endptr = strchr(packet, '*');
    len = (int)endptr - (int)&packet[5];
    strncpy(buf, &packet[5], len);
    buf[len + 1] = '\0';
    // Angefragter Datensatz
    set_number = atoi(buf);        

    addr = sizeof(tPositionSet) * set_number;
    ee24xx_read_bytes( addr, sizeof(tPositionSet), (uint8_t*) &Pos);
    uputs("$"); 
    // Umwandeln der Werte in Zeichen mit fhrenden Nullen
    sprintf(buf, "%010lu",  Pos.date);
    uputs(buf);
    sprintf(buf, "%010lu",  Pos.time);
    uputs(buf);
    sprintf(buf, "%010lu",  Pos.latitude);
    uputs(buf);
    sprintf(buf, "%010lu",  Pos.longitude);
    uputs(buf);
    sprintf(buf, "%04d", Pos.altitude);
    uputs(buf);
    uputchar(Pos.latitude_ns);
    uputchar(Pos.longitude_ew);    
    uputs("*\r\n");     
    // Interrupts wieder an
    sei();
}
//************************************************************************************************************
void pcTransWREQ(void) {
    char buf[10];
    itoa (NumWaypoints, buf, 10);
    uputs("$WCNT,");
    uputs(buf);
    uputs("*\r\n");    
}
//************************************************************************************************************
void pcTransWSET(uint8_t* packet) {
    uint16_t set_number, len;
    uint16_t wpt;
    char *endptr;
    char buf[15];    
    // Interrupts aus
    cli();
    // Buffer leeren, da ansonsten mehrere Befehle mehrmals abgearbeitet werden
    bufferFlush(&UartRxBuffer);
    wpt = 
    buf[0] = 0;
    endptr = strchr(packet, '*');
    len = (int)endptr - (int)&packet[5];
    strncpy(buf, &packet[5], len);
    buf[len + 1] = '\0';
    // Angefragter Datensatz
    set_number = atoi(buf);        
    // Auslesen der Daten aus dem EEPROM    
    wpt = eeprom_read_word(&eeWaypoint[set_number]);
    uputs("$"); 
    // Umwandeln der Werte in Zeichen mit fhrenden Nullen
    sprintf(buf, "%05u",  wpt);
    uputs(buf);
    uputs("*\r\n");     
    // Interrupts wieder an
    sei();
}
void pcTransCLEAR(void) {
    // Index auf 0 setzen damit neue GPS-Daten die alten berschreiben

}

